/*
 * Node.java
 *
 * Created on 13. listopad 2006, 15:00
 *
 * @author Jiri Kremser
 * @version 1.0
 * @windowtitle class Node
 */

/* indexing of coordinates (system) :
 y : value of position[]
 x : index of position[]
 
y
8 |
7 |
6 |
5 |
4 |
3 |
2 |
1 |
   -----------------
    0 1 2 3 4 5 6 7 x   position[]  ~~  x - 1
*/


package dam;
    import java.util.ArrayList;
    import java.util.Arrays;


/**
 * class Main
 */
public class Node {
    
    public static int N = 8;
    
    private int[] position = new int[N];
    private Node prev;
    private int[] queenConflict = new int[N];
    private ArrayList bestMoves;
    public static int[][] visited = new int[100][Node.N];
    public static int visitedCount = 0;

//*****************************************************************************
        
    /**
     * Constructor  <Node (int[], Node)>
     *
     * @param position chessboard
     * @param prev previous node 
     */
    public Node(int[] position, Node prev) throws Exception{        
        Exception badDimensionException = new Exception();
        if (position.length != Node.N) {
            throw badDimensionException;
        }
        for (int i = 0; i < Node.N; i++) {
            this.position[i] = position[i];
        }        
        this.prev = prev;
        this.setQueensC();
    }
    
    
    /**
     * Constructor 2  <Node (Node, Node)>
     *
     * @param node actual node
     * @param prev previous node
     */
    public Node(Node node, Node prev) {        
        this.position = node.position; 
        this.prev = prev;
        this.setQueensC();
    }
    

    /**
     * Node getPrev()
     *
     * @return previous node
     */
    public Node getPrev() {
        if (prev == null) {
            return null;
        }
        else {
            return new Node(prev,prev.getPrev());
        }        
    }
    
        
    /**
     * int getC(int,int)
     *
     * @param x x-coordinate
     * @param y y-coordinate
     * @return number of queen-conflicts at the particular place
     */
    private int getC(int x,int y) {
        int c = 0;

        for (int i = 0 ; i < Node.N; i++) {
            if (this.position[i] == y) {
                c++;
            }        
            if (i != Node.N - 1) { // ! last loop
                if ( ((x + i < Node.N) && (this.position[x + i] == (y + 1 + i)))
                || ((x + i < Node.N) && (this.position[x + i] == (y - 1 - i)))
                || ((x - 2 - i >= 0) && (this.position[x - i - 2] == (y - 1 - i)))
                || ((x - 2 - i >= 0) && (this.position[x - i - 2] == (y + 1 + i))) ){
                    c++;
                }
            }
        }
        return c;
    }
   
    
    /**
     * int getQueenC(int)
     *
     * @param queen number of the queen
     * @return number of queen-conflicts of the particular queen
     */
    private int getQueenC(int queen) {
        return this.getC(queen,position[queen-1]) - 1;
    }
    
    
    /**
     * int getTotalQueenC()
     *
     * @return number of queen-conflicts (sum)
     */
    private int getTotalQueenC() {
        int sum = 0;
        for (int i = 0; i < Node.N; i++) {
            sum += this.queenConflict[i];
        }
        return sum;
    }
    
    
    /**
     * setQueensC()
     *
     */
    private void setQueensC() {
        for (int i = 0; i < Node.N; i++) {
            this.queenConflict[i] = this.getC(i + 1, this.position[i]) - 1;
        }        
    }
    
            
    /**
     * ArrayList getMaxQueenC()
     *
     * @return ordinary numbers of most conflict queens
     */
    private ArrayList getMaxQueenC() {
        int max = 0, akt;        
        ArrayList minQueenConflicts = new ArrayList();        
        for (int i = 0; i < Node.N; i++) {
            if (this.queenConflict[i] > max ) {
                max = this.queenConflict[i];
            }
        }
        for (int j = 0; j < Node.N; j++) {
            if (max == this.queenConflict[j]) {
                minQueenConflicts.add(j+1);
            }
        }
        return minQueenConflicts;
    }    


    /**
     * ArrayList getMinCmoves(int)
     *
     * @param queen number of the queen
     * @return y-coordinates of minimal conflict moves of the particular queen
     */
    private ArrayList getMinCmoves(int queen) {
        int min = Node.N, akt;
        int[] conflicts = new int [Node.N];
        ArrayList minConflicts = new ArrayList();
        
        for (int i = 0; i < Node.N ; i++) {
            if (i != (position[queen-1] - 1)) {
                conflicts[i] = this.getC(queen,i+1);                
                if (min > conflicts[i]) {
                    min = conflicts[i];
                }
            }
        }

        for (int j = 0; j < Node.N; j++) {
            if ( (j != (position[queen-1] - 1)) && (conflicts[j] == min) 
                && (!this.move(queen,j+1).isVisited())) {                
                minConflicts.add(j+1);
            }
        }
        return minConflicts;
    }
    
    
    /**
     * Node move(int,int)
     *
     * @param queen number of the queen
     * @param HPosition y-coordinate (destination)
     * @return result of movement (new node)
     */
    private Node move(int queen, int HPosition) {
        Node nod = null;
        try {
            nod = new Node(this.position,this.prev);
        }
        catch (Exception bde){

        }
        nod.position[queen - 1] = HPosition;
        nod.setQueensC();
        return nod;
    }
    
    
    /**
     * visit()
     */
    public void visit() {
        Node.visited[Node.visitedCount] = this.position;
        visitedCount++;
    }
    
    
    /**
     * boolean isVisited()
     *
     * @return true if node is visited
     */
    private boolean isVisited() {
        for (int i = 0;i<Node.visitedCount;i++) {
            if (Arrays.equals(Node.visited[i],this.position)) {
                return true;
            }
        }
        return false;
    }

    
   /**
     * cleanVisited()
     */
    public static void cleanVisited() {        
        for (int i = 0; i < Node.visitedCount; i++) {
            Node.visited[i][0] = -1;
        }
        Node.visitedCount = 0;
    }
      
    
    /**
     * Node selectBest()
     * 
     * Selects one of the best moves
     * 
     * @return one of the best moves
     * @throws java.lang.Exception 
     */
    public Node selectBest() throws Exception{
        ArrayList listQueens = this.getMaxQueenC();
        ArrayList bestMoves = new ArrayList();
        int queen,y;
        int nodeTotalQueenC =  this.getTotalQueenC();
        int min = nodeTotalQueenC;
        int aktTotalQueenC;
        
        for (int i = 0; i < listQueens.size(); i++) {
            queen = (Integer)listQueens.get(i);
            ArrayList listMoves = new ArrayList();
            listMoves = this.getMinCmoves(queen);
            for (int j = 0; j < listMoves.size(); j++) {
                y = (Integer) listMoves.get(j);
                this.move(queen,y).setQueensC();
                aktTotalQueenC = this.move(queen,y).getTotalQueenC();
                if (aktTotalQueenC <= nodeTotalQueenC) {
                    if (aktTotalQueenC < min) {
                        min = aktTotalQueenC;
                    }                                        
                }
            }
        }
        for (int i = 0; i < listQueens.size(); i++) {
            queen = (Integer)listQueens.get(i);
            ArrayList listMoves = this.getMinCmoves(queen);
            for (int j = 0; j < listMoves.size(); j++) {
                y = (Integer)listMoves.get(j);
                aktTotalQueenC = this.move(queen,y).getTotalQueenC();
                if ((min != (Node.N * Node.N)) &&  (aktTotalQueenC == min)) {
                    Object node = this.move(queen,y);
                    bestMoves.add( node);
                }
            }
        }        
        if  (bestMoves.size() == 0) { // no possible moves
            Exception noSolutionException = new Exception();
            throw noSolutionException;
        }
        if (min <= nodeTotalQueenC) {
            int ind = ( (int) ( Math.random() * 1000) % (bestMoves.size()) );
            return (Node) bestMoves.get(ind);
        }
        else return null;
    }
    
    
    /**
     * boolean isEnd()
     *
     * @return true if the problem has been solved
     */
    public boolean isEnd() {
        if (this.getTotalQueenC() == 0) {
            return true;
        }
        else return false;
    }
    
    
    /**
     * String toString()
     */    
    public String toString() {
        boolean eval = false;
        String board = "   ---------------------";
        for (int h = 0; h < Node.N - 5; h++) {
            board += "----";
        }
        board += "\n";
        for (int i = Node.N; i > 0; i--) {
            for (int j = 0; j < Node.N; j++) {
                if (j == 0) {
                    board += i + " " + ((i >= 10)?"":" ");
                }
                if (position[j] == i) {
                    board += "| Q ";
                }
                else board += "| " + ((eval==true)?this.getC(j + 1,i):" ") + " ";                
            }
            board += "|\n   ---------------------";
            for (int j = 0; j < Node.N - 5; j++) {
                board += "----";
            }
            board += "\n";
        }
        board += "     A   B   C   D   E";
        for (int h = 0; h < Node.N - 5; h++) {
            board += "   " + (char) ((int) 'F' + h) ;
        }
        
        return board;
    }
}
